###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Base utility functions to load models and project-wise environment variables."""
import logging
import os
from pathlib import Path

from transformers import (AutoImageProcessor, AutoModel, AutoModelForSeq2SeqLM,
                          AutoTokenizer, VisionEncoderDecoderModel)

logger = logging.getLogger('ocr.general')

root = Path(os.environ.get('TRANSFORMERS_CACHE', '.'))
logger.debug(f'Cache dir: {root}')
dev = os.environ.get('DEVICE', 'cpu')

def load(loader, model_id: str):
    """Use the specified loader to load a transformers specific Class."""
    try:
        mid = root / model_id
        logger.debug(f'Attempt loading from store: "{loader}" "{mid}"')
        res = loader.from_pretrained(mid)
    except Exception:
        # Needed to catch some weird exception from transformers
        # eg: huggingface_hub.utils._validators.HFValidationError: Repo id must use alphanumeric chars or
        # '-', '_', '.', '--' and '..' are forbidden, '-' and '.' cannot start or end the name, max length is 96: ...
        logger.debug(f'Attempt loading from cache: "{loader}" "{model_id}" "{root}"')
        res = loader.from_pretrained(model_id, cache_dir=root)
    return res

mapping = {
    'tokenizer': AutoTokenizer,
    'ved_model': VisionEncoderDecoderModel,
    'model': AutoModel,
    'image_processor': AutoImageProcessor,
    'seq2seq': AutoModelForSeq2SeqLM
}

accept_device = ['ved_model', 'seq2seq', 'model']

def load_hugginface_model(model_id: str, request: list[str]) -> list:
    """Load the requested HuggingFace's Classes for the model into the memory of the globally specified device.

    Args:
        model_id (str): The HuggingFace model id to load, or a path to a local model.
        request (list[str]): A list of HuggingFace's Classes to load.

    Raises:
        ValueError: If the model_id is not found or if the requested Class is not supported.

    Returns:
        _type_: A list of the requested Classes.
    """    """"""
    res = {}
    for r in request:
        if r not in mapping:
            raise ValueError(f'Unknown request: {r}')
        cls = load(mapping[r], model_id)
        if cls is None:
            raise ValueError(f'Could not load model: {model_id}')

        if r in accept_device:
            cls = cls.to(dev)

        res[r] = cls

    return res



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Functions and piplines to perform Box OCR on an image."""
import logging
from typing import Hashable, Iterable

import easyocr
import numpy as np
import torch
from PIL import Image

from .. import models as m
from ..queues import box_queue as q
from .huggingface import dev

READER = None

logger = logging.getLogger('ocr.general')

BOX_MODEL_ID = None
BBOX_MODEL_OBJ = None

def unload_box_model():
    """Remove the current box model from memory."""
    global BBOX_MODEL_OBJ, READER, BOX_MODEL_ID

    logger.info(f'Unloading BOX model: {BOX_MODEL_ID}')
    if BOX_MODEL_ID == 'easyocr':
        pass
    READER = None
    BBOX_MODEL_OBJ = None
    BOX_MODEL_ID = None

    if dev == 'cuda':
        torch.cuda.empty_cache()


def load_box_model(model_id: str):
    """Load a box model into memory."""
    global BBOX_MODEL_OBJ, READER, BOX_MODEL_ID

    if BOX_MODEL_ID == model_id:
        return

    logger.info(f'Loading BOX model: {model_id}')
    if model_id == 'easyocr':
        READER = easyocr.Reader([], gpu=(dev == 'cuda'), recognizer=False)
        BBOX_MODEL_OBJ, _ = m.OCRBoxModel.objects.get_or_create(name=model_id)
        BOX_MODEL_ID = model_id
    else:
        raise NotImplementedError

    logger.debug(f'OCR model loaded: {model_id}')
    logger.debug(f'OCR model object: {BBOX_MODEL_OBJ}')

def get_box_model() -> m.OCRBoxModel:
    """Get the current box model."""
    return BBOX_MODEL_OBJ

def intersections(bboxes: Iterable[tuple[int, int, int, int]], margin: int = 5) -> list[set[int]]:
    """Determine the intersections between a list of bounding boxes.

    Args:
        bboxes (Iterable[tuple[int, int, int, int]]): List of bounding boxes in lrbt format.
        margin (int, optional): Number of extra pixels outside of the boxes that define an intersection. Defaults to 5.

    Returns:
        list[set[int]]: List of sets of indexes of the boxes that intersect.
    """
    res = []

    for i,(l1,r1,b1,t1) in enumerate(bboxes):
        l1 -= margin
        r1 += margin
        b1 -= margin
        t1 += margin

        for j,(l2,r2,b2,t2) in enumerate(bboxes):
            if i == j:
                continue

            if l1 >= r2 or r1 <= l2 or b1 >= t2 or t1 <= b2:
                continue

            for ptr in res:
                if i in ptr or j in ptr:
                    break
            else:
                ptr = set()
                res.append(ptr)

            ptr.add(i)
            ptr.add(j)

    return res

def merge_bboxes(bboxes: Iterable[tuple[int, int, int, int]]) -> list[tuple[int, int, int, int]]:
    """Merge a list of intersecting bounding boxes. All intersecting boxes are merged into a single box.

    Args:
        bboxes (Iterable[Iterable[int]]): Iterable of bounding boxes in lrbt format.

    Returns:
        list[tuple[int]]: List of merged bounding boxes in lrbt format.
    """
    res = []
    bboxes = np.array(bboxes)
    inters = intersections(bboxes)

    lst = list(range(len(bboxes)))

    torm = set()
    for app in inters:
        app = list(app)
        data = bboxes[app].reshape(-1,4)
        l = data[:,0].min()
        r = data[:,1].max()
        b = data[:,2].min()
        t = data[:,3].max()

        res.append([l,b,r,t])

        torm = torm.union(app)

    for i in lst:
        if i in torm:
            continue
        l,r,b,t = bboxes[i]
        res.append([l,b,r,t])

    return res

def _box_pipeline(image: Image.Image, options: dict = None) -> list[tuple[int, int, int, int]]:
    """Perform box OCR on an image.

    Args:
        image (Image.Image): A Pillow image on which to perform OCR.
        options (dict, optional): A dictionary of options.

    Raises:
        NotImplementedError: The type of model specified is not implemented.

    Returns:
        list[tuple[int, int, int, int]]: A list of bounding boxes in lrbt format.
    """

    if options is None:
        options = {}

    # reader.recognize(image)
    if BOX_MODEL_ID == 'easyocr':
        image = image.convert('RGB')
        results = READER.detect(np.array(image))

        # Axis rectangles
        bboxes = results[0][0]

        # Free (NOT IMPLEMENTED)
        # ...

        bboxes = merge_bboxes(bboxes)
    else:
        raise NotImplementedError

    return bboxes

def box_pipeline(*args, id_: Hashable, block: bool = True, **kwargs):
    """Queue a box OCR pipeline.

    Args:
        id_ (Hashable): A unique identifier for the OCR task.
        block (bool, optional): Whether to block until the task is complete. Defaults to True.

    Returns:
        Union[str, Message]: The text extracted from the image (block=True) or a Message object (block=False).
    """
    msg = q.put(
        id_ = id_,
        msg = {'args': args, 'kwargs': kwargs},
        handler = _box_pipeline,
    )

    if block:
        return msg.response()
    return msg

def box_run(
        img_obj: m.Image, lang: m.Language, image: Image.Image = None,
        force: bool = False, options: m.OptionDict = None
        ) -> list[m.BBox]:
    """High level function to perform box OCR on an image. Will attempt to reuse a previous run if possible.

    Args:
        img_obj (m.Image): An Image object from the database.
        lang (m.Language): A Language object from the database (not every model is gonna use this).
        image (Image.Image, optional): The Pillow image to be used for OCR if a previous result is not found.
            Defaults to None.
        force (bool, optional): If true, re-run the OCR even if a previous result is found. Defaults to False.
        options (m.OptionDict, optional): An OptionDict object from the database. Defaults to None.

    Raises:
        ValueError: ValueError is raised if at any step of the pipeline an image is required but not provided.

    Returns:
        list[m.BBox]: A list of BBox objects containing the resulting bounding boxes.
    """
    options_obj = options or m.OptionDict.objects.get(options={})
    params = {
        'image': img_obj,
        'model': BBOX_MODEL_OBJ,
        'options': options_obj,
        'lang_src': lang,
    }

    bbox_run = m.OCRBoxRun.objects.filter(**params).first()
    if bbox_run is None or force:
        if image is None:
            raise ValueError('Image is required for BBox OCR')
        logger.info('Running BBox OCR')
        opt_dct = options_obj.options
        bboxes = box_pipeline(
            image,
            id_=img_obj.md5,
            options=opt_dct,
            )
        # Create it here to avoid having a failed entry in DB
        bbox_run = m.OCRBoxRun.objects.create(**params)
        for bbox in bboxes:
            l,b,r,t = bbox
            m.BBox.objects.create(
                l=l,
                b=b,
                r=r,
                t=t,
                image=img_obj,
                from_ocr=bbox_run,
                )
    else:
        logger.info(f'Reusing BBox OCR <{bbox_run.id}>')
    logger.info(f'BBox OCR result: {len(bbox_run.result.all())} boxes')

    return list(bbox_run.result.all())



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Full OCR + translation pipelines."""
import logging

from PIL import Image

from .. import models as m
from .box import box_run
from .lang import get_lang_dst, get_lang_src
from .ocr import ocr_run
from .tsl import tsl_run

logger = logging.getLogger('ocr.general')

def ocr_tsl_pipeline_lazy(md5: str, options: dict = None) -> list[dict]:
    """
    Try to lazily generate reponse from md5.
    Should raise a ValueError if the operation is not possible (fails at any step).
    """
    if options is None:
        options = {}
    logger.debug(f'LAZY: START {md5}')
    res = []
    try:
        img_obj= m.Image.objects.get(md5=md5)
    except m.Image.DoesNotExist as exc:
        raise ValueError(f'Image with md5 {md5} does not exist') from exc
    bbox_obj_list = box_run(img_obj, get_lang_src())
    for bbox_obj in bbox_obj_list:
        text_obj = ocr_run(bbox_obj, get_lang_src())
        text_obj = next(text_obj)
        tsl_obj = tsl_run(text_obj, get_lang_src(), get_lang_dst(), lazy=True)
        tsl_obj = next(tsl_obj)

        text = text_obj.text
        new = tsl_obj.text

        res.append({
            'ocr': text,
            'tsl': new,
            'box': bbox_obj.lbrt,
            })

    logger.debug('LAZY: DONE')
    return res

# This is already kinda lazy, but the idea for the lazy version is to
# check if all results are available just with the md5, and if not,
# ask the extension to send the binary to minimize traffic
def ocr_tsl_pipeline_work(img: Image.Image, md5: str, force: bool = False, options: dict = None) -> list[dict]:
    """
    Generate response from md5 and binary.
    Will attempt to behave lazily at every step unless force is True.
    """
    if options is None:
        options = {}
    logger.debug(f'WORK: START {md5}')
    res = []

    img_obj, _ = m.Image.objects.get_or_create(md5=md5)
    bbox_obj_list = box_run(img_obj, get_lang_src() ,image=img)

    texts = []
    for bbox_obj in bbox_obj_list:
        logger.debug(str(bbox_obj))

        text_obj = ocr_run(bbox_obj, get_lang_src(), image=img, force=force, block=False)
        next(text_obj)
        texts.append(text_obj)

    texts = [next(_) for _ in texts]
    logger.debug(f'OCR DONE: {texts}')

    trans = []
    for text_obj in texts:
        tsl_obj = tsl_run(text_obj, get_lang_src(), get_lang_dst(), force=force, block=False)
        next(tsl_obj)
        trans.append(tsl_obj)

    trans = [next(_) for _ in trans]
    logger.debug(f'TRANSLATION DONE: {trans}')

    for bbox_obj, text_obj, tsl_obj in zip(bbox_obj_list, texts, trans):
        text = text_obj.text
        new = tsl_obj.text

        res.append({
            'ocr': text,
            'tsl': new,
            'box': bbox_obj.lbrt,
            })

    logger.debug('WORK: DONE')
    return res



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Functions and piplines to perform OCR on an image."""
import logging
from typing import Generator, Hashable, Union

import torch
from PIL import Image

from .. import models as m
from ..messaging import Message
from ..queues import ocr_queue as q
from .huggingface import dev, load_hugginface_model
from .tesseract import tesseract_pipeline

logger = logging.getLogger('ocr.general')

OBJ_MODEL_ID: str = None
OCR_MODEL = None
OCR_TOKENIZER = None
OCR_IMAGE_PROCESSOR = None

OCR_MODEL_OBJ: m.OCRModel = None

NO_SPACE_LANGUAGES = [
    'ja', 'zh', 'lo', 'my'
]

def unload_ocr_model():
    """Remove the current OCR model from memory."""
    global OCR_MODEL_OBJ, OCR_MODEL, OCR_TOKENIZER, OCR_IMAGE_PROCESSOR, OBJ_MODEL_ID

    logger.info(f'Unloading OCR model: {OBJ_MODEL_ID}')
    OCR_MODEL = None
    OCR_TOKENIZER = None
    OCR_IMAGE_PROCESSOR = None
    OCR_MODEL_OBJ = None
    OBJ_MODEL_ID = None

    if dev == 'cuda':
        torch.cuda.empty_cache()

def load_ocr_model(model_id: str):
    """Load an OCR model into memory."""
    global OCR_MODEL_OBJ, OCR_MODEL, OCR_TOKENIZER, OCR_IMAGE_PROCESSOR, OBJ_MODEL_ID

    if OBJ_MODEL_ID == model_id:
        return

    # mid = root / model_id
    logger.info(f'Loading OCR model: {model_id}')
    if model_id == 'tesseract':
        pass
    else:
        res = load_hugginface_model(model_id, request=['ved_model', 'tokenizer', 'image_processor'])
        OCR_MODEL = res['ved_model']
        OCR_TOKENIZER = res['tokenizer']
        OCR_IMAGE_PROCESSOR = res['image_processor']

    OCR_MODEL_OBJ, _ = m.OCRModel.objects.get_or_create(name=model_id)
    OBJ_MODEL_ID = model_id

    logger.debug(f'OCR model loaded: {model_id}')
    logger.debug(f'OCR model object: {OCR_MODEL_OBJ}')

def get_ocr_model() -> m.OCRModel:
    """Return the current OCR model."""
    return OCR_MODEL_OBJ

def _ocr(img: Image.Image, lang: str = None, bbox: tuple[int, int, int, int] = None, options: dict = None) -> str:
    """Perform OCR on an image.

    Args:
        img (Image.Image):  A Pillow image on which to perform OCR.
        lang (str, optional): The language to use for OCR. (Not every model will use this)
        bbox (tuple[int, int, int, int], optional): The bounding box of the text on the image in lbrt format.
        options (dict, optional): A dictionary of options to pass to the OCR model.

    Raises:
        TypeError: If img is not a Pillow image.

    Returns:
        str: The text extracted from the image.
    """
    if options is None:
        options = {}
    if not isinstance(img, Image.Image):
        raise TypeError(f'img should be PIL Image, but got {type(img)}')
    img = img.convert('RGB')

    if bbox:
        img = img.crop(bbox)

    if OBJ_MODEL_ID == 'tesseract':
        generated_text = tesseract_pipeline(img, lang)
    else:
        pixel_values = OCR_IMAGE_PROCESSOR(img, return_tensors='pt').pixel_values
        if dev == 'cuda':
            pixel_values = pixel_values.cuda()
        generated_ids = OCR_MODEL.generate(pixel_values)
        generated_text = OCR_TOKENIZER.batch_decode(generated_ids, skip_special_tokens=True)[0]

    return generated_text

def ocr(*args, id_: Hashable, block: bool = True, **kwargs) -> Union[str, Message]:
    """Queue a text OCR pipeline.

    Args:
        id_ (Hashable): A unique identifier for the OCR task.
        block (bool, optional): Whether to block until the task is complete. Defaults to True.

    Returns:
        Union[str, Message]: The text extracted from the image (block=True) or a Message object (block=False).
    """
    msg = q.put(
        id_ = id_,
        msg = {'args': args, 'kwargs': kwargs},
        handler = _ocr,
    )

    if block:
        return msg.response()
    return msg

def ocr_run(
        bbox_obj: m.BBox, lang: m.Language,  image: Image.Image = None, options: m.OptionDict = None,
        force: bool = False, block: bool = True,
        ) -> Generator[Union[Message, m.Text], None, None]:
    """High level function to perform OCR on an image.

    Args:
        bbox_obj (m.BBox): The BBox object from the database.
        lang (m.Language): The Language object from the database.
        image (Image.Image, optional): The image on which to perform OCR. Needed if no previous OCR run exists, or
            force is True.
        options (m.OptionDict, optional): The OptionDict object from the database containing the options for the OCR.
        force (bool, optional): Whether to force the OCR to run again even if a previous run exists. Defaults to False.
        block (bool, optional): Whether to block until the task is complete. Defaults to True.

    Raises:
        ValueError: ValueError is raised if at any step of the pipeline an image is required but not provided.

    Yields:
        Generator[Union[Message, m.Text], None, None]:
            If block is True, yields a Message object for the OCR run first and the resulting Text object second.
            If block is False, yields the resulting Text object.
    """
    options_obj = options
    if options_obj is None:
        options_obj = m.OptionDict.objects.get(options={})
    params = {
        'bbox': bbox_obj,
        'model': OCR_MODEL_OBJ,
        'lang_src': lang,
        'options': options_obj,
    }
    ocr_run_obj = m.OCRRun.objects.filter(**params).first()
    if ocr_run_obj is None or force:
        if image is None:
            raise ValueError('Image is required for OCR')
        logger.info('Running OCR')

        id_ = (bbox_obj.id, OCR_MODEL_OBJ.id, lang.id)
        mlang = getattr(lang, OCR_MODEL_OBJ.language_format or 'iso1')
        opt_dct = options_obj.options
        text = ocr(
            image,
            lang=mlang,
            bbox=bbox_obj.lbrt,
            options=opt_dct,
            id_=id_,
            block=block,
            )
        if not block:
            yield text
            text = text.response()
        if lang.iso1 in NO_SPACE_LANGUAGES:
            text = text.replace(' ', '')
        text_obj, _ = m.Text.objects.get_or_create(
            text=text,
            )
        params['result'] = text_obj
        ocr_run_obj = m.OCRRun.objects.create(**params)
    else:
        if not block:
            # Both branches should have the same number of yields
            yield None
        logger.info(f'Reusing OCR <{ocr_run_obj.id}>')
        text_obj = ocr_run_obj.result
        # text = ocr_run.result.text

    yield text_obj



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django models for the ocr_translate app."""
from django.db import models

LANG_LENGTH = 32

class OptionDict(models.Model):
    """Dictionary of options for OCR and translation"""
    options = models.JSONField(unique=True)

    def __str__(self):
        return str(self.options)

class Language(models.Model):
    """Language used for translation"""
    name = models.CharField(max_length=64, unique=True)
    iso1 = models.CharField(max_length=8, unique=True)
    iso2b = models.CharField(max_length=8, unique=True)
    iso2t = models.CharField(max_length=8, unique=True)
    iso3 = models.CharField(max_length=32, unique=True)

    easyocr = models.CharField(max_length=32, null=True)
    tesseract = models.CharField(max_length=32, null=True)
    facebookM2M = models.CharField(max_length=32, null=True)

    break_chars = models.CharField(max_length=512, null=True)
    ignore_chars = models.CharField(max_length=512, null=True)

    def __str__(self):
        return str(self.iso1)

class OCRModel(models.Model):
    """OCR model using hugging space naming convention"""
    name = models.CharField(max_length=128)
    languages = models.ManyToManyField(Language, related_name='ocr_models')

    language_format = models.CharField(max_length=32, null=True)

    def __str__(self):
        return str(self.name)

class OCRBoxModel(models.Model):
    """OCR model for bounding boxes"""
    name = models.CharField(max_length=128)
    languages = models.ManyToManyField(Language, related_name='box_models')

    language_format = models.CharField(max_length=32, null=True)

    def __str__(self):
        return str(self.name)

class TSLModel(models.Model):
    """Translation models using hugging space naming convention"""
    name = models.CharField(max_length=128)
    src_languages = models.ManyToManyField(Language, related_name='tsl_models_src')
    dst_languages = models.ManyToManyField(Language, related_name='tsl_models_dst')

    language_format = models.CharField(max_length=32, null=True)

    def __str__(self):
        return str(self.name)

class Image(models.Model):
    """Image registered as the md5 of the uploaded file"""
    md5 = models.CharField(max_length=32, unique=True)

class BBox(models.Model):
    """Bounding box of a text in an image"""
    l = models.IntegerField(null=False)
    b = models.IntegerField(null=False)
    r = models.IntegerField(null=False)
    t = models.IntegerField(null=False)

    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='bboxes')
    from_ocr = models.ForeignKey('OCRBoxRun', on_delete=models.CASCADE, related_name='result')

    @property
    def lbrt(self):
        """Return the bounding box as a tuple of (left, bottom, right, top)"""
        return self.l, self.b, self.r, self.t

    def __str__(self):
        return f'{self.lbrt}'

class Text(models.Model):
    """Text extracted from an image or translated from another text"""
    text = models.TextField()
    # lang = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='texts')

class OCRBoxRun(models.Model):
    """OCR run on an image using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='ocr_box_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='box_src')

    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='to_box')
    model = models.ForeignKey(OCRBoxModel, on_delete=models.CASCADE, related_name='box_runs')
    # result = models.ForeignKey(BBox, on_delete=models.CASCADE, related_name='from_ocr')

class OCRRun(models.Model):
    """OCR run on an image using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='ocr_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='ocr_src')

    # image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='to_ocr')
    bbox = models.ForeignKey(BBox, on_delete=models.CASCADE, related_name='to_ocr')
    model = models.ForeignKey(OCRModel, on_delete=models.CASCADE, related_name='ocr_runs')
    result = models.ForeignKey(Text, on_delete=models.CASCADE, related_name='from_ocr')

class TranslationRun(models.Model):
    """Translation run on a text using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='trans_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='trans_src')
    lang_dst = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='trans_dst')

    text = models.ForeignKey(Text, on_delete=models.CASCADE, related_name='to_trans')
    model = models.ForeignKey(TSLModel, on_delete=models.CASCADE, related_name='tsl_runs')
    result = models.ForeignKey(Text, on_delete=models.CASCADE, related_name='from_trans')



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Functions and piplines to perform OCR on images using tesseract."""
import logging
import os
from pathlib import Path

import requests
from PIL import Image
from pytesseract import Output, image_to_string

from .huggingface import root

logger = logging.getLogger('ocr.general')

MODEL_URL = 'https://github.com/tesseract-ocr/tessdata_best/raw/main/{}.traineddata'

DATA_DIR = Path(os.getenv('TESSERACT_PREFIX', root / 'tesseract'))

VERTICAL_LANGS = ['jpn', 'chi_tra', 'chi_sim', 'kor']

DOWNLOAD = os.getenv('TESSERACT_ALLOW_DOWNLOAD', 'false').lower() == 'true'
CONFIG = False

def download_model(lang: str):
    """Download a tesseract model for a given language.

    Args:
        lang (str): A language code for tesseract.

    Raises:
        ValueError: If the model could not be downloaded.
    """
    if not DOWNLOAD:
        raise ValueError('TESSERACT_ALLOW_DOWNLOAD is false. Downloading models is not allowed')
    create_config()

    logger.info(f'Downloading tesseract model for language {lang}')
    dst = DATA_DIR / f'{lang}.traineddata'
    if dst.exists():
        return
    res = requests.get(MODEL_URL.format(lang), timeout=5)
    if res.status_code != 200:
        raise ValueError(f'Could not download model for language {lang}')

    with open(DATA_DIR / f'{lang}.traineddata', 'wb') as f:
        f.write(res.content)

    if lang in VERTICAL_LANGS:
        download_model(lang + '_vert')

def create_config():
    """Create a tesseract config file. Run only once"""
    global CONFIG
    if CONFIG:
        return
    CONFIG = True

    logger.info('Creating tesseract tsv config')
    cfg = DATA_DIR / 'configs'
    cfg.mkdir(exist_ok=True, parents=True)

    dst = cfg / 'tsv'
    if dst.exists():
        return
    with dst.open('w') as f:
        f.write('tessedit_create_tsv 1')

# Page segmentation modes:
#   0    Orientation and script detection (OSD) only.
#   1    Automatic page segmentation with OSD.
#   2    Automatic page segmentation, but no OSD, or OCR.
#   3    Fully automatic page segmentation, but no OSD. (Default)
#   4    Assume a single column of text of variable sizes.
#   5    Assume a single uniform block of vertically aligned text.
#   6    Assume a single uniform block of text.
#   7    Treat the image as a single text line.
#   8    Treat the image as a single word.
#   9    Treat the image as a single word in a circle.
#  10    Treat the image as a single character.
#  11    Sparse text. Find as much text as possible in no particular order.
#  12    Sparse text with OSD.
#  13    Raw line. Treat the image as a single text line,
#        bypassing hacks that are Tesseract-specific.
def tesseract_pipeline(img: Image.Image, lang: str, favor_vertical: bool = True) -> str:
    """Run tesseract on an image.

    Args:
        img (Image.Image): An image to run tesseract on.
        lang (str): A language code for tesseract.
        favor_vertical (bool, optional): Wether to favor vertical or horizontal configuration for languages that
            can be written vertically. Defaults to True.

    Returns:
        str: The text extracted from the image.
    """    """"""
    create_config()
    if not (DATA_DIR / f'{lang}.traineddata').exists():
        download_model(lang)
    logger.info(f'Running tesseract for language {lang}')

    psm = 6
    if lang in VERTICAL_LANGS:
        exp = 1 if favor_vertical else -1
        if img.height * 1.5**exp > img.width:
            psm = 5

    # Using image_to_string will atleast preserve spaces
    res = image_to_string(
        img,
        lang=lang,
        config=f'--tessdata-dir {DATA_DIR.as_posix()} --psm {psm}',
        output_type=Output.DICT
        )

    return res['text']



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Admin interface for ocr_translate app."""
from django.contrib import admin

from . import models as m


class LanguageAdmin(admin.ModelAdmin):
    """Admin interface for Language model"""
    list_display = ('name', 'iso1', 'iso2b', 'iso2t', 'iso3')

class OCRBoxModelAdmin(admin.ModelAdmin):
    """Admin interface for OCRBoxModel model"""
    list_display = ('name',)

class OCRModelAdmin(admin.ModelAdmin):
    """Admin interface for OCRModel model"""
    list_display = ('name',)

class TSLModelAdmin(admin.ModelAdmin):
    """Admin interface for TSLModel model"""
    list_display = ('name',)


admin.site.register(m.Language, LanguageAdmin)
admin.site.register(m.OCRBoxModel, OCRBoxModelAdmin)
admin.site.register(m.OCRModel, OCRModelAdmin)
admin.site.register(m.TSLModel, TSLModelAdmin)



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Messaging and worker+queue system for ocr_translate."""
import logging
import queue
import threading
import time
from typing import Callable, Hashable, Iterable, Union

logger = logging.getLogger('ocr.worker')


class NotHandled():
    """Dummy object to be used as default response of an unresolved message."""

class Message():
    """Message object to be used in WorkerMessageQueue. This class is used to send a message in the queue, but
    also allow the sending function to wait for the response of the message."""
    NotHandled = NotHandled
    def __init__(
            self, id_: Hashable, msg: dict, handler: Callable,
            batch_args: tuple = (), batch_kwargs: Iterable = ()
            ):
        """Message object to be used in WorkerMessageQueue.

        Args:
            id_ (Hashable): Message id. Used to identify messages with the same id.
            msg (dict): Message to be passed to the handler.
            handler (Callable): Handler function to be called with the message.
            batch_args (tuple, optional): Indexes of the args to be batched. Defaults to ().
            batch_kwargs (Iterable, optional): Keys of the kwargs to be batched. Defaults to ().
        """
        self.id_ = id_
        self.msg = msg
        self.handler = handler
        self.batch_args = batch_args
        self.batch_kwargs = batch_kwargs
        self._response = NotHandled

    def resolve(self):
        """Resolve the message by calling the handler with the message.
        This operation is synchronous and will block the exection until the handler is done."""
        try:
            self._response = self.handler(*self.msg.get('args', ()), **self.msg.get('kwargs', {}))
        except Exception as exc:
            self._response = exc
            raise
        logger.debug(f'MSG Resolved {self.msg} -> {self._response}')

        # Make sure to dereference the message to avoid keeping raw images in memory
        # since i am gonna keep the message in the queue after it is resolved (for msg caching)
        del self.msg

    def batch_resolve(self, others: Iterable['Message']):
        """Resolve multiple messages with one call to the handler.
        The handler must be able to handle the specified batched args and kwargs,
        as either the expected type or a list of the expected type.
        The handler must return a list of the same length as the number of messages to be resolved, with the same order.
        """
        logger.debug(f'MSG Batch Resolving {self.msg} with {len(others)} other messages')
        # Check if these checks are necessary (maybe just let the handler fail)
        # Main problem would be running messages with different non batched args that produce worng results
        # But having all these checks might slow down the batching too much
        if any(_.handler != self.handler for _ in others):
            raise ValueError('All messages must have the same handler')
        check = len(self.msg['args'])
        if any(len(_.msg['args']) != check for _ in others):
            raise ValueError('All messages must have the same number of args')
        check = set(self.msg['kwargs'].keys())
        if any(set(_.msg['kwargs'].keys()) != check for _ in others):
            raise ValueError('All messages must have the same kwargs')
        # Should also check that the non-batched args and kwargs are the same for all messages

        args = [[a] if i in self.batch_args else a for i, a in enumerate(self.msg['args'])]
        kwargs = {k:[v] if k in self.batch_kwargs else v for k, v in self.msg['kwargs'].items()}

        for msg in others:
            for i in self.batch_args:
                args[i].append(msg.msg['args'][i])
            for k in self.batch_kwargs:
                kwargs[k].append(msg.msg['kwargs'][k])

        respones = self.handler(*args, **kwargs)
        for msg, r in zip([self, *others], respones):
            logger.debug(f'MSG Batch Resolved {msg.msg} -> {r}')
            msg.set_response(r)

            # Make sure to dereference the message to avoid keeping raw images in memory
            # since i am gonna keep the message in the queue after it is resolved (for msg caching)
            del msg.msg

    @property
    def is_resolved(self) -> bool:
        """Whether the message has been resolved or not."""
        return self._response is not NotHandled

    def set_response(self, response):
        """Set the response of the message."""
        self._response = response

    def response(self, timeout: float = 0, poll: float = 0.2):
        """Get the response of the message.

        Args:
            timeout (float, optional): Timeout in seconds to wait for the message to be resolved.
                Defaults to 0 (no timeout).
            poll (float, optional): Polling interval in seconds. Defaults to 0.2.

        Raises:
            TimeoutError: If the message is not resolved after the timeout.

        Returns:
            Any: The response of the message (return value of the handler called on the msg content).
        """
        start = time.time()

        while not self.is_resolved:

            if time.time() - start > timeout > 0:
                raise TimeoutError('Message resolution timed out')
            time.sleep(poll)

        return self._response

    def __repr__(self):
        return f'Message({self.msg}), Handler: {self.handler.__name__}'

    def __str__(self):
        return f'Message({self.msg}), Handler: {self.handler.__name__}'

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Message):
            return False

        res = True
        res &= self.id_ == __value.id_
        res &= self.msg == __value.msg
        res &= self.handler == __value.handler
        return res

    def copy(self) -> 'Message':
        """Return a copy of the message. Used for tests."""
        return Message(
            self.id_, dict(self.msg), self.handler,
            batch_args=self.batch_args, batch_kwargs=self.batch_kwargs
            )

class MessageQueue(queue.SimpleQueue):
    """Message queue with worker threads to resolve messages. This class extends queue.SimpleQueue, by adding:
        - Message caching/reuse (When a new message with the same id is put in the queue, the old one is returned)
        - Message batching (Messages with the same batch_id are grouped together and resolved with one handler call)
    """
    def __init__(
            self,
            *args,
            reuse_msg: bool = True,
            # max_len: int = 0,
            allow_batching: bool = False,
            batch_timeout: float = 0.5,
            batch_args: tuple = (), batch_kwargs: Iterable = (),
            **kwargs
            ):
        """Create a new WorkerMessageQueue.

        Args:
            reuse_msg (bool, optional): Whether to reuse messages with the same id. Defaults to True.
            max_len (int, optional): Max number of messages in cache before starting to remove solved messages
                from cache. Defaults to 0 (no limit).
            allow_batching (bool, optional): Whether to allow batching of messages. Defaults to False.
            batch_timeout (float, optional): Timeout for batching. When get is called, wait `timeout` seconds for other
                incoming messages. Defaults to 0.5.
            batch_args (tuple, optional): Indexes of the args to be batched. Defaults to ().
            batch_kwargs (Iterable, optional): Keys of the kwargs to be batched. Defaults to ().
        """
        super().__init__(*args, **kwargs)
        self.registered = {}
        self.batch_pools = {}
        self.msg_to_batch_pool = {}
        self.batch_resolve_flagged = []
        self.reuse_msg = reuse_msg
        # self.max_len = max_len
        if allow_batching:
            if len(batch_args) == 0 and len(batch_kwargs) == 0:
                raise ValueError('At least one batch arg or kwarg must be specified with batching enabled.')
        self.allow_batching = allow_batching
        self.batch_timeout = batch_timeout
        self.batch_args = batch_args
        self.batch_kwargs = batch_kwargs

    def put(self, id_: Hashable, msg: dict, handler: Callable, batch_id: Hashable = None) -> Message:
        """Put a new message in the queue.

        Args:
            id_ (Hashable): Id of the message. Used to identify messages with the same id.
            msg (dict): Message to be passed to the handler.
            handler (Callable): Handler function to be called with the message.
            batch_id (Hashable, optional): Id of the batch to which the message belongs. Defaults to None.

        Raises:
            NotImplementedError: If the max_len is reached.

        Returns:
            Message: The message object.
        """
        if batch_id is not None and not self.allow_batching:
            raise ValueError('Batching is not allowed')

        if self.reuse_msg and id_ in self.registered:
            logger.debug(f'Reusing message {id_}')
            return self.registered[id_]
        # if self.max_len > 0 and self.qsize() > self.max_len:
        #     # Remove solved messages from cache
        #     # Only 1by1 or all?
        #     raise NotImplementedError('Max len reached')

        res = Message(id_, msg, handler, batch_args=self.batch_args, batch_kwargs=self.batch_kwargs)
        if self.allow_batching and batch_id is not None:
            self.msg_to_batch_pool[id_] = batch_id
            ptr = self.batch_pools.setdefault(batch_id, [])
            ptr.append(res)

        if self.reuse_msg:
            self.registered[id_] = res

        super().put(res)

        return res

    def get(self, *args, **kwargs) -> Union[Message, list[Message]]:
        """Get a message or list of messages from the queue.

        Returns:
            Union[Message, list[Message]]: A message or a list of messages, depending on whether batching is enabled.
        """
        msg = super().get(*args, **kwargs)
        while msg.id_ in self.batch_resolve_flagged:
            self.batch_resolve_flagged.remove(msg.id_)
            msg = super().get(*args, **kwargs)

        if self.allow_batching and msg.id_ in self.msg_to_batch_pool:
            # Wait for more messages to come
            logger.debug(f'Batching message {msg.id_}')
            time.sleep(self.batch_timeout)

            logger.debug(f'Batching message {msg.id_} done')
            pool_id = self.msg_to_batch_pool[msg.id_]
            logger.debug(f'Batching message {msg.id_} pool id {pool_id}')
            pool = self.batch_pools.pop(pool_id)
            logger.debug(f'Batching message {msg.id_} pool {pool}')
            for msg in pool:
                self.msg_to_batch_pool.pop(msg.id_)
                self.batch_resolve_flagged.append(msg.id_)
            return pool

        return msg

    def get_msg(self, msg_id: str):
        """Get a message from the cache. If the message is not in the cache, return None.

        Args:
            msg_id (str): Id of the message.

        Returns:
            _type_: The message object or None.
        """
        if not self.reuse_msg:
            raise ValueError('Message caching is disabled')

        return self.registered.get(msg_id, None)

class Worker():
    """Worker object to be used in WorkerMessageQueue."""
    def __init__(self, attached_queue: MessageQueue, poll_interval: float = .2):
        self.queue = attached_queue
        self.kill = False
        self.running = False
        self.thread = None
        self.poll_interval = poll_interval

    def _worker(self):
        """Worker function that consumes messages from the queue and resolves them."""
        self.running = True
        while not self.kill:
            try:
                msg = self.queue.get(timeout=self.poll_interval)
            except queue.Empty:
                continue
            logger.debug(f'Worker consuming {msg}')
            if isinstance(msg, list):
                if len(msg) == 1:
                    msg[0].resolve()
                else:
                    msg[0].batch_resolve(msg[1:])
            else:
                msg.resolve()

        self.running = False

    def start(self):
        """Start the worker thread."""
        self.kill = False # Allow restarting the worker
        self.thread = threading.Thread(target=self._worker, daemon=True)
        self.thread.start()

    def stop(self):
        """Stop the worker thread."""
        self.kill = True
        self.thread.join()

class WorkerMessageQueue():
    """Bundle together the queue and its workers."""
    def __init__(self, *args, num_workers: int = 1, **kwargs):
        """Create a new WorkerMessageQueue.

        Args:
            num_workers (int, optional): Number of workers to spawn. Defaults to 1.
        """
        self.msg_queue = MessageQueue(*args, **kwargs)
        self.workers = [Worker(self.msg_queue) for _ in range(num_workers)]

    def put(self, id_: Hashable, msg: dict, handler: Callable, batch_id: Hashable = None) -> Message:
        """Call the put method of the queue."""
        return self.msg_queue.put(id_, msg, handler, batch_id=batch_id)

    def get(self, *args, **kwargs) -> Union[Message, list[Message]]:
        """Call the get method of the queue."""
        return self.msg_queue.get(*args, **kwargs)

    def get_msg(self, msg_id: str):
        """Call the get_msg method of the queue."""
        return self.msg_queue.get_msg(msg_id)

    def start_workers(self):
        """Start all the worker threads registered to this queue."""
        for worker in self.workers:
            worker.start()

    def stop_workers(self):
        """Stop all the worker threads registered to this queue."""
        for worker in self.workers:
            worker.stop()



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Functions and piplines to perform translation on text."""
import logging
import re
from typing import Generator, Hashable, Union

import torch
from transformers import M2M100Tokenizer

from .. import models as m
from ..messaging import Message
from ..queues import tsl_queue as q
from .huggingface import dev, load_hugginface_model

logger = logging.getLogger('ocr.general')

TSL_MODEL_ID = None
TSL_MODEL = None
TSL_TOKENIZER = None
TSL_MODEL_OBJ = None

def unload_tsl_model():
    """Remove the current TSL model from memory."""
    global TSL_MODEL_OBJ, TSL_MODEL, TSL_TOKENIZER, TSL_MODEL_ID

    logger.info(f'Unloading TSL model: {TSL_MODEL_ID}')
    TSL_MODEL = None
    TSL_TOKENIZER = None
    TSL_MODEL_OBJ = None
    TSL_MODEL_ID = None

    if dev == 'cuda':
        torch.cuda.empty_cache()

def load_tsl_model(model_id):
    """Load a TSL model into memory."""
    global TSL_MODEL_OBJ, TSL_MODEL, TSL_TOKENIZER, TSL_MODEL_ID

    if TSL_MODEL_ID == model_id:
        return

    logger.info(f'Loading TSL model: {model_id}')
    res = load_hugginface_model(model_id, request=['seq2seq', 'tokenizer'])
    TSL_MODEL = res['seq2seq']
    TSL_TOKENIZER = res['tokenizer']

    TSL_MODEL_OBJ, _ = m.TSLModel.objects.get_or_create(name=model_id)
    TSL_MODEL_ID = model_id

def get_tsl_model() -> m.TSLModel:
    """Get the current TSL model."""
    return TSL_MODEL_OBJ

def pre_tokenize(
        text: str,
        ignore_chars: str = None, break_chars: str = None, break_newlines: bool = True
        ) -> list[str]:
    """Pre-tokenize a text string.

    Args:
        text (str): Text to tokenize.
        ignore_chars (str, optional): String of characters to ignore. Defaults to None.
        break_chars (str, optional): String of characters to break on. Defaults to None.
        break_newlines (bool, optional): Whether to break on newlines. Defaults to True.

    Returns:
        list[str]: List of string tokens.
    """
    if ignore_chars is not None:
        text = re.sub(f'[{ignore_chars}]+', '', text)
    if break_chars is None:
        break_chars = ''
    if break_newlines:
        break_chars += '\n'

    break_chars = re.escape(break_chars)
    tokens = text
    if len(break_chars) > 0:
        tokens = re.split(f'[{break_chars}+]', text)

    if isinstance(tokens, str):
        tokens = [text]

    res = list(filter(None, tokens))
    return res if len(res) > 0 else [' ']

def get_mnt(ntok: int, options: dict) -> int:
    """Get the maximum number of new tokens to generate."""
    min_max_new_tokens = options.get('min_max_new_tokens', 20)
    max_max_new_tokens = options.get('max_max_new_tokens', 512)
    max_new_tokens = options.get('max_new_tokens', 20)
    max_new_tokens_ratio = options.get('max_new_tokens_ratio', 3)

    if min_max_new_tokens > max_max_new_tokens:
        raise ValueError('min_max_new_tokens must be less than max_max_new_tokens')

    mnt = min(
        max_max_new_tokens,
        max(
            min_max_new_tokens,
            max_new_tokens,
            max_new_tokens_ratio * ntok
        )
    )
    return mnt

def _tsl_pipeline(
        text: Union[str,list[str]],
        lang_src: str, lang_dst: str,
        options: dict = None
        ) -> Union[str,list[str]]:
    """Translate a text using a TSL model.

    Args:
        text (Union[str,list[str]]): Text to translate. Can be batched to a list of strings.
        lang_src (str): Source language.
        lang_dst (str): Destination language.
        options (dict, optional): Options for the translation. Defaults to {}.

    Raises:
        TypeError: If text is not a string or a list of strings.

    Returns:
        Union[str,list[str]]: Translated text. If text is a list, returns a list of translated strings.
    """
    if options is None:
        options = {}
    TSL_TOKENIZER.src_lang = lang_src

    break_newlines = options.get('break_newlines', True)
    break_chars = options.get('break_chars', None)
    ignore_chars = options.get('ignore_chars', None)

    args = (ignore_chars, break_chars, break_newlines)
    if isinstance(text, list):
        tokens = [pre_tokenize(t, *args) for t in text]
    elif isinstance(text, str):
        tokens = pre_tokenize(text, *args)
    else:
        raise TypeError(f'Unsupported type for text: {type(text)}')

    logger.debug(f'TSL: {tokens}')
    if len(tokens) == 0:
        return ''
    encoded = TSL_TOKENIZER(
        tokens,
        return_tensors='pt',
        padding=True,
        truncation=True,
        is_split_into_words=True
        )
    ntok = encoded['input_ids'].shape[1]
    encoded.to(dev)

    mnt = get_mnt(ntok, options)

    kwargs = {
        'max_new_tokens': mnt,
    }
    if isinstance(TSL_TOKENIZER, M2M100Tokenizer):
        kwargs['forced_bos_token_id'] = TSL_TOKENIZER.get_lang_id(lang_dst)

    logger.debug(f'TSL ENCODED: {encoded}')
    logger.debug(f'TSL KWARGS: {kwargs}')
    generated_tokens = TSL_MODEL.generate(
        **encoded,
        **kwargs,
        )

    tsl = TSL_TOKENIZER.batch_decode(generated_tokens, skip_special_tokens=True)
    logger.debug(f'TSL: {tsl}')

    if isinstance(text, str):
        tsl = tsl[0]
    return tsl

def tsl_pipeline(*args, id_: Hashable, batch_id: Hashable = None, block: bool = True, **kwargs):
    """Queue a text translation pipeline.

    Args:
        id_ (Hashable): A unique identifier for the OCR task.
        block (bool, optional): Whether to block until the task is complete. Defaults to True.

    Returns:
        Union[str, Message]: The text extracted from the image (block=True) or a Message object (block=False).
    """
    msg = q.put(
        id_ = id_,
        batch_id = batch_id,
        msg = {'args': args, 'kwargs': kwargs},
        handler = _tsl_pipeline,
    )

    if block:
        return msg.response()
    return msg

def tsl_run(
        text_obj: m.Text, src: m.Language, dst: m.Language, options: m.OptionDict = None,
        force: bool = False,
        block: bool = True,
        lazy: bool = False
        ) -> Generator[Union[Message, m.Text], None, None]:
    """Run a TSL pipeline on a text object.

    Args:
        text_obj (m.Text): Text object from the database to translate.
        src (m.Language): Source language object from the database.
        dst (m.Language): Destination language object from the database.
        options (m.OptionDict, optional): OptionDict object from the database. Defaults to None.
        force (bool, optional): Whether to force a new TSL run. Defaults to False.
        block (bool, optional): Whether to block until the task is complete. Defaults to True.
        lazy (bool, optional): Whether to raise an error if the TSL run is not found. Defaults to False.

    Raises:
        ValueError: If lazy and force are both True or if lazy is True and the TSL run is not found.

    Yields:
        Generator[Union[Message, m.Text], None, None]:
            If block is True, yields a Message object for the TSL run first and the resulting Text object second.
            If block is False, yields the resulting Text object.
    """
    if lazy and force:
        raise ValueError('Cannot force + lazy TSL run')
    model_obj = get_tsl_model()
    options_obj = options or m.OptionDict.objects.get(options={})
    params = {
        'options': options_obj,
        'text': text_obj,
        'model': model_obj,
        'lang_src': src,
        'lang_dst': dst,
    }
    tsl_run_obj = m.TranslationRun.objects.filter(**params).first()
    if tsl_run_obj is None or force:
        if lazy:
            raise ValueError('Value not found for lazy TSL run')
        logger.info('Running TSL')
        id_ = (text_obj.id, model_obj.id, options_obj.id, src.id, dst.id)
        batch_id = (model_obj.id, options_obj.id, src.id, dst.id)
        opt_dct = options_obj.options
        opt_dct.setdefault('break_chars', src.break_chars)
        opt_dct.setdefault('ignore_chars', src.ignore_chars)
        new = tsl_pipeline(
            text_obj.text,
            getattr(src, model_obj.language_format),
            getattr(dst, model_obj.language_format),
            options=opt_dct,
            id_=id_,
            batch_id=batch_id,
            block=block,
            )
        if not block:
            yield new
            new = new.response()
        text_obj, _ = m.Text.objects.get_or_create(
            text = new,
            )
        params['result'] = text_obj
        tsl_run_obj = m.TranslationRun.objects.create(**params)
    else:
        if not block:
            # Both branches should have the same number of yields
            yield None
        logger.info(f'Reusing TSL <{tsl_run_obj.id}>')
        # new = tsl_run_obj.result.text

    yield tsl_run_obj.result
